# 
# namcap rules - pathdepends
# Copyright (C) 2016 Kyle Keen <keenerd@gmail.com>
# 
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
# 

"""
This contains a collection of essentially one-line rules:
If a certain path is detected then a certain dependency is expected.

Anything fancier than this should get its own rule.
"""

import re
from Namcap.ruleclass import *

class PathDependsRule(TarballRule):
	name = "pathdepends"
	description = "Check for simple implicit path dependencies"
	# list of path regex, dep name, reason tag
	subrules = [
	{'path': '^usr/share/glib-2\.0/schemas$',
		'dep':'dconf',
		'reason':'dconf-needed-for-glib-schemas'},
	{'path': '^usr/lib/gio/modules/.*\.so$',
		'dep':'glib2',
		'reason':'glib2-needed-for-gio-modules'},
	{'path': '^usr/share/icons/hicolor$',
		'dep':'hicolor-icon-theme',
		'reason':'hicolor-icon-theme-needed-for-hicolor-dir'},
	{'path': '^usr/share/mime$',
		'dep':'shared-mime-info',
		'reason':'shared-mime-info-needed'},
	]
	def analyze(self, pkginfo, tar):
		names = [entry.name for entry in tar]
		for subrule in self.subrules:
			pattern = re.compile(subrule['path'])
			if any(pattern.search(n) for n in names):
				dep = subrule['dep']
				reason = subrule['reason']
				pkginfo.detected_deps[dep].append( (reason, ()) )

# vim: set ts=4 sw=4 noet:



# -*- coding: utf-8 -*-
# 
# namcap rules - shebangdepends
# Copyright (C) 2003-2009 Jason Chu <jason@archlinux.org>
# Copyright (C) 2011 Rémy Oudompheng <remy@archlinux.org>
# 
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
# 

"""Checks dependencies on programs specified in shebangs."""

import shutil
import Namcap.package
from Namcap.util import is_script, script_type
from Namcap.ruleclass import *

def scanshebangs(fileobj, filename, scripts):
	"""
	Scan a file for shebang and stores the interpreter name.

	Stores
	  scripts -- a dictionary { program => set(scripts) }
	"""

	# test magic bytes
	if not is_script(fileobj):
		return
	# process shebang line
	cmd = script_type(fileobj)
	if cmd != None:
		assert(isinstance(cmd, str))
		scripts.setdefault(cmd, set()).add(filename)

def findowners(scriptlist):
	"""
	Find owners for executables.

	Returns:
	  pkglist -- a dictionary { package => set(programs) }
	  orphans -- a set of scripts not found
	"""

	pkglist = {}
	scriptfound = set()

	for s in scriptlist:
		out = shutil.which(s)
		if not out:
			continue

		# strip leading slash
		scriptpath = out.lstrip('/')
		for pkg in Namcap.package.get_installed_packages():
			pkg_files = [fname for fname, fsize, fmode in pkg.files]
			if scriptpath in pkg_files:
				pkglist.setdefault(pkg.name, set()).add(s)
				scriptfound.add(s)

	orphans = list(set(scriptlist) - scriptfound)
	return pkglist, orphans

class ShebangDependsRule(TarballRule):
	name = "shebangdepends"
	description = "Checks dependencies semi-smartly."
	def analyze(self, pkginfo, tar):
		scriptlist = {}

		for entry in tar:
			if not entry.isfile():
				continue
			f = tar.extractfile(entry)
			scanshebangs(f, entry.name, scriptlist)
			f.close()

		# find packages owning interpreters
		pkglist, orphans = findowners(scriptlist)
		for dep, progs in pkglist.items():
			if not isinstance(progs, set):
				continue
			needing = set().union(*[scriptlist[prog] for prog in progs])
			reasons = pkginfo.detected_deps.setdefault(dep, [])
			reasons.append((
				'programs-needed %s %s',
				(str(list(progs)), str(list(needing)))
				))

		# Do the script handling stuff
		for i, v in scriptlist.items():
			files = list(v)
			self.infos.append(("script-link-detected %s in %s", (i, str(files))))

		# Handle "no package associated" errors
		self.warnings.extend([("library-no-package-associated %s", i)
			for i in orphans])

		# Check for packages in testing
		for i in scriptlist:
			p = Namcap.package.load_testing_package(i)
			q = Namcap.package.load_from_db(i)
			if p != None and q != None and p["version"] == q["version"]:
				self.warnings.append(("dependency-is-testing-release %s", i))

# vim: set ts=4 sw=4 noet:



#
# namcap rules - systemdlocation
# Copyright (C) 2015 James Harvey <jamespharvey20@gmail.com>
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

import os
from Namcap.ruleclass import *

class systemdlocationRule(TarballRule):
	name = "systemdlocation"
	description = "Checks for systemd files in /etc/systemd/system/"
	def analyze(self, pkginfo, tar):
		# don't have this warning for the systemd package
		if 'name' in pkginfo:
			if pkginfo['name'] == 'systemd':
				return
		# don't have this warning for packages that provides systemd
		if 'provides' in pkginfo:
			if 'systemd' in pkginfo['provides']:
				return
		for entry in tar:
			# ignore the actual directory, as that's handled by emptydirs
			if entry.isdir():
				continue;

			name = os.path.normpath(entry.name)

			# check for files in /etc/systemd/system/
			if name.startswith('etc/systemd/system/'):
				self.warnings.append(("systemd-location %s", name))

# vim: set ts=4 sw=4 noet:



# -*- coding: utf-8 -*-
#
# namcap tests - rpath
# Copyright (C) 2011 Rémy Oudompheng <remy@archlinux.org>
# 
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
#   USA
# 

import os
from Namcap.tests.makepkg import MakepkgTest
import Namcap.rules.rpath

class RpathTest(MakepkgTest):
	pkgbuild = """
pkgname=__namcap_test_rpath
pkgver=1.0
pkgrel=1
pkgdesc="A package"
arch=('i686' 'x86_64')
url="http://www.example.com/"
license=('GPL')
depends=('glibc')
source=()
options=(!purge !zipman)
build() {
  cd "${srcdir}"
  echo "int main() { return 0; }" > main.c
  /usr/bin/gcc -o main.o -c main.c
  /usr/bin/ld -o main --rpath /home/evil/lib main.o -lc
}
package() {
  install -D -m 755 "$srcdir/main" "$pkgdir/usr/bin/evilprogram"
}
"""
	def test_rpath_files(self):
		"Package with fancy RPATHs"
		pkgfile = "__namcap_test_rpath-1.0-1-%(arch)s.pkg.tar" % { "arch": self.arch }
		with open(os.path.join(self.tmpdir, "PKGBUILD"), "w") as f:
			f.write(self.pkgbuild)
		self.run_makepkg()
		pkg, r = self.run_rule_on_tarball(
				os.path.join(self.tmpdir, pkgfile),
				Namcap.rules.rpath.package
				)
		self.assertEqual(r.errors, [
			("insecure-rpath %s %s",
			("/home/evil/lib", "usr/bin/evilprogram"))
		])
		self.assertEqual(r.warnings, [])
		self.assertEqual(r.infos, [])

# vim: set ts=4 sw=4 noet:




# -*- coding: utf-8 -*-
#
# namcap rules - javafiles
# Copyright (C) 2011 Rémy Oudompheng <remy@archlinux.org>
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

from Namcap.ruleclass import *
from Namcap.util import is_java

class JavaFiles(TarballRule):
	name = "javafiles"
	description = "Check for existence of Java classes or JARs"
	def analyze(self, pkginfo, tar):
		has_java_classes = False
		javas = []
		for entry in tar:
			# is it a regular file ?
			if not entry.isfile():
				continue
			# is it a JAR file ?
			if entry.name.endswith('.jar'):
				javas.append(entry.name)
				#self.infos.append( ('jar-file-found %s', entry.name) )
				continue
			# is it a CLASS file ?
			f = tar.extractfile(entry)
			if is_java(f):
				javas.append(entry.name)
				#self.infos.append( ('java-class-file-found %s', entry.name) )
			f.close()
		if javas:
			reasons = pkginfo.detected_deps.setdefault('java-runtime', [])
			reasons.append( ('java-runtime-needed %s', ', '.join(javas)) )

# vim: set ts=4 sw=4 noet:



# -*- coding: utf-8 -*-
#
# namcap tests - tests for the version module
# Copyright (C) 2015 Rikard Falkeborn <rikard.falkeborn@gmail.com>
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
#   USA
#

import os
import unittest
import re
import Namcap.version


class VersionTests(unittest.TestCase):
	def test_manpage(self):
		''' Test that the manpage and program has the same version.'''
		here = os.path.dirname(os.path.realpath(__file__))
		with open(os.path.join(here, '..', '..', 'namcap.1')) as f:
			first_line = f.readline()
		match = re.search('"namcap (.*?)"', first_line)
		self.assertEqual(match.group(1), Namcap.version.get_version())

# vim: set ts=4 sw=4 noet:



# 
# namcap rules - externalhooks
# Copyright (C) 2016 Kyle Keen <keenerd@gmail.com>
# 
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
# 

from Namcap.ruleclass import *

class ExternalHooksRule(TarballRule):
	name = "externalhooks"
	description = "Check the .INSTALL for commands covered by hooks"
	hooked = [
		'update-desktop-database',
		'update-mime-database',
		'install-info',
		'glib-compile-schemes',
		'gtk-update-icon-cache',
		'xdg-icon-resource',
		'gconfpkg',
		'gio-querymodules',
		'fc-cache',
		'mkfontscale',
		'mkfontdir',
		'systemd-sysusers',
		'systemd-tmpfiles',
		'vlc-cache-gen',
	]
	def analyze(self, pkginfo, tar):
		if ".INSTALL" not in tar.getnames():
			return
		f = tar.extractfile(".INSTALL")
		text = f.read().decode('utf-8', 'ignore')
		f.close()
		for command in self.hooked:
			if command in text:
				self.warnings = [('external-hooks-warning', ())]
				self.infos.append(('external-hooks-name %s', command))

# vim: set ts=4 sw=4 noet:



# namcap tests - shebangdepends
# Copyright (C) 2016 Kyle Keen <keenerd@gmail.com>
# 
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
#   USA
# 

import os
from Namcap.tests.makepkg import MakepkgTest
import Namcap.rules.shebangdepends

class ShebangDependsTest(MakepkgTest):
	pkgbuild = """
pkgname=__namcap_test_shebangdepends
pkgver=1.0
pkgrel=1
pkgdesc="A package"
arch=('any')
url="http://www.example.com/"
license=('GPL')
depends=()
source=()
options=(!purge !zipman)
build() {
  cd "${srcdir}"
  echo -e "#! /usr/bin/env python\nprint('a script')" > python_sample
  echo -e "#!/bin\\xffary/da\\x00ta\ncrash?" > binary_sample
}
package() {
  install -Dm755 "$srcdir/python_sample" "$pkgdir/usr/bin/python_sample"
  install -Dm755 "$srcdir/binary_sample" "$pkgdir/usr/share/binary_sample"
}
"""
	def test_shebangdepends(self):
		"Package with missing python dependency"
		pkgfile = "__namcap_test_shebangdepends-1.0-1-any.pkg.tar"
		with open(os.path.join(self.tmpdir, "PKGBUILD"), "w") as f:
			f.write(self.pkgbuild)
		self.run_makepkg()
		pkg, r = self.run_rule_on_tarball(
				os.path.join(self.tmpdir, pkgfile),
				Namcap.rules.shebangdepends.ShebangDependsRule
				)
		e, w, i = Namcap.depends.analyze_depends(pkg)
		self.assertEqual(e, [
			('dependency-detected-not-included %s (%s)',
				('python', "programs ['python'] needed in scripts ['usr/bin/python_sample']"))
		])
		self.assertEqual(w, [])

# vim: set ts=4 sw=4 noet:




# 
# namcap tests - externalhooks rule
# Copyright (C) 2016 Kyle Keen <keenerd@gmail.com>
# 
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
# 

import os
from Namcap.tests.makepkg import MakepkgTest
import Namcap.rules.externalhooks

class ExternalHooksTest(MakepkgTest):
	pkgbuild_generic = """
pkgname=__namcap_test_externalhooks
pkgver=1.0
pkgrel=1
pkgdesc="A package"
arch=('any')
url="http://www.example.com/"
license=('GPL')
depends=('glibc')
options=(emptydirs)
install=sample.install
source=()
build() {
  true
}
package() {
  true
}

"""
	install_bad = """
post_install() {
  update-desktop-database -q
  gtk-update-icon-cache -q -t -f usr/share/icons/hicolor
}
"""
	install_good = """
post_install() {
  true
}
"""
	def test_bad_install(self):
		pkgfile = "__namcap_test_externalhooks-1.0-1-any.pkg.tar"
		with open(os.path.join(self.tmpdir, "PKGBUILD"), "w") as f:
			f.write(self.pkgbuild_generic)
		with open(os.path.join(self.tmpdir, "sample.install"), "w") as f:
			f.write(self.install_bad)
		self.run_makepkg()
		pkg, r = self.run_rule_on_tarball(
				os.path.join(self.tmpdir, pkgfile),
				Namcap.rules.externalhooks.ExternalHooksRule
				)
		self.assertEqual(r.errors, [])
		self.assertEqual(r.warnings, [('external-hooks-warning', ())])
		self.assertEqual(r.infos,
			[('external-hooks-name %s', 'update-desktop-database'),
			('external-hooks-name %s', 'gtk-update-icon-cache')])
	def test_good_install(self):
		pkgfile = "__namcap_test_externalhooks-1.0-1-any.pkg.tar"
		with open(os.path.join(self.tmpdir, "PKGBUILD"), "w") as f:
			f.write(self.pkgbuild_generic)
		with open(os.path.join(self.tmpdir, "sample.install"), "w") as f:
			f.write(self.install_good)
		self.run_makepkg()
		pkg, r = self.run_rule_on_tarball(
				os.path.join(self.tmpdir, pkgfile),
				Namcap.rules.externalhooks.ExternalHooksRule
				)
		self.assertEqual(r.errors, [])
		self.assertEqual(r.warnings, [])
		self.assertEqual(r.infos, [])
    
# vim: set ts=4 sw=4 noet:



# 
# namcap rules - mimefiles
# Copyright (C) 2009 Hugo Doria <hugo@archlinux.org>
# Copyright (C) 2011 Rémy Oudompheng <remy@archlinux.org>
# 
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
# 

from Namcap.ruleclass import *

class MimeDesktopRule(TarballRule):
	name = "mimedesktop"
	description = "Check for MIME desktop file depends"
	def analyze(self, pkginfo, tar):
		for entry in tar:
			if entry.issym():
				continue
			if not entry.name.startswith("usr/share/applications"):
				continue
			if not entry.name.endswith(".desktop"):
				continue
			with tar.extractfile(entry) as f:
				if not any(l.startswith(b"MimeType=") for l in f):
					continue
				pkginfo.detected_deps["desktop-file-utils"].append( ('desktop-file-utils-needed', ()) )
				break

# vim: set ts=4 sw=4 noet:



# 
# namcap rules - carch
# Copyright (C) 2003-2009 Jason Chu <jason@archlinux.org>
# 
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
# 

"""Verifies that no specific host type is used"""

import re
from Namcap.ruleclass import *

class package(PkgbuildRule):
	name = "carch"
	description = "Verifies that no specific host type is used"
	def analyze(self, pkginfo, tar):
		arches = ['arm', 'i586', 'i686', 'ppc', 'x86_64']
		archmatch = re.compile(r'\b(%s)\b' % '|'.join(arches))
		# Match an arch=(i686) line
		archline = re.compile(r'\w*arch=')
		for line in pkginfo.pkgbuild:
			# strip any comments
			line = line.split('#')[0]
			match = archmatch.search(line)
			if not match:
				continue
			if archline.match(line):
				continue
			if '$CARCH' in line:
				continue
			if '${CARCH}' in line:
				continue
			self.warnings.append(("specific-host-type-used %s", match.group(1)))

# vim: set ts=4 sw=4 noet:



# -*- coding: utf-8 -*-
# 
# namcap rules - depends
# Copyright (C) 2003-2009 Jason Chu <jason@archlinux.org>
# Copyright (C) 2011 Rémy Oudompheng <remy@archlinux.org>
# 
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
# 

"""Checks dependencies semi-smartly."""

import re
from Namcap.ruleclass import *
import Namcap.tags
from Namcap import package

# manual overrides for relationships outside the normal metadata
implicit_provides = {
	'java-environment': ['java-runtime'],
}

def single_covered(depend):
	"Returns full coverage tree of one package, with loops broken"
	covered = set()
	todo = set([depend])
	while todo:
		i = todo.pop()
		covered.add(i)
		pac = package.load_from_db(i)
		if pac is None:
			continue
		todo |= set(pac["depends"]) - covered

	return covered - set([depend])

def getcovered(dependlist):
	"""
	Returns full coverage tree set, without packages
	from self-loops (iterable of package names)
	"""

	covered = set()
	for d in dependlist:
		covered |= single_covered(d)
	return covered

def getcustom(pkginfo):
	custom_name = {'^mingw-': ['mingw-w64-crt'],}
	custom_depend = set()
	for pattern in custom_name:
		if re.search(pattern, pkginfo['name']):
			custom_depend.update(custom_name[pattern])
	return custom_depend

def getprovides(depends):
	provides = {}
	for i in depends:
		provides[i] = set()
		if i in implicit_provides:
			provides[i].update(implicit_provides[i])
		pac = package.load_from_db(i)
		if pac is None:
			continue
		if not pac["provides"]:
			continue
		provides[i].update(pac["provides"])
	return provides

def analyze_depends(pkginfo):
	errors, warnings, infos = [], [], []

	# compute needed dependencies + recursive
	dependlist = set(pkginfo.detected_deps.keys())
	indirectdependlist = getcovered(dependlist)
	customlist = getcustom(pkginfo)
	for i in indirectdependlist:
		infos.append(("dependency-covered-by-link-dependence %s", i))
	needed_depend = dependlist | indirectdependlist | customlist
	# the minimal set of needed dependencies
	smartdepend = set(dependlist) - indirectdependlist

	# Find all the covered dependencies from the PKGBUILD
	pkginfo.setdefault("depends", [])
	explicitdepend = set(pkginfo["depends"])
	implicitdepend = getcovered(explicitdepend)
	pkgbuild_depend = explicitdepend | implicitdepend

	# Include the optdepends from the PKGBUILD
	pkginfo.setdefault("optdepends", [])
	optdepend = set(pkginfo["optdepends"])
	optdepend |= getcovered(optdepend)

	# Get the provides so we can reference them later
	# smartprovides : depend => (packages provided by depend)
	smartprovides = getprovides(smartdepend | explicitdepend)

	# The set of all provides for detected dependencies
	allprovides = set()
	for plist in smartprovides.values():
		allprovides |= plist

	# Do the actual message outputting stuff
	for i in smartdepend:
		# if the needed package is itself:
		if i == pkginfo["name"]:
			continue
		# if the dependency is satisfied
		if i in pkgbuild_depend:
			continue
		# if the dependency is pulled as a provider for some explicit dep
		if smartprovides[i] & pkgbuild_depend:
			continue
		# the dependency is satisfied by a provides
		if i in allprovides:
			continue
		# compute dependency reason
		reasons = pkginfo.detected_deps[i]
		reason_strings = [Namcap.tags.format_message(reason) for reason in reasons]
		reason = ', '.join(reason_strings)
		# still not found, maybe it is specified as optional
		if i in optdepend:
			warnings.append(("dependency-detected-but-optional %s (%s)", (i, reason)))
			continue
		# maybe, it is pulled as a provider for an optdepend
		if smartprovides[i] & optdepend:
			warnings.append(("dependency-detected-but-optional %s (%s)", (i, reason)))
			continue
		# now i'm pretty sure i didn't find it.
		errors.append(("dependency-detected-not-included %s (%s)", (i, reason)))

	for i in pkginfo["depends"]:
		# multilib packages usually depend on their regular counterparts
		if pkginfo["name"].startswith('lib32-') and i == pkginfo["name"].partition('-')[2]:
			continue
		# the package provides an important dependency
		if i in smartprovides and (smartprovides[i] & smartdepend):
			continue
		# a needed dependency is superfluous it is implicitly satisfied
		if i in implicitdepend and (i in smartdepend or i in indirectdependlist):
			warnings.append(("dependency-already-satisfied %s", i))
		# a dependency is unneeded if:
		#   it is not in the depends as we see them
		#   it does not pull some needed dependency which provides it
		elif i not in needed_depend and i not in allprovides:
			warnings.append(("dependency-not-needed %s", i))
	infos.append(("depends-by-namcap-sight depends=(%s)", ' '.join(smartdepend) ))

	return errors, warnings, infos

# vim: set ts=4 sw=4 noet:



#
# namcap tests - glibfiles
# Copyright (C) 2011 Rémy Oudompheng <remy@archlinux.org>
# Copyright (C) 2016 Kyle Keen <keenerd@gmail.com>
# 
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
#   USA
# 

import os
from Namcap.tests.makepkg import MakepkgTest
import Namcap.rules.pathdepends

class PathDependsTest(MakepkgTest):
	pkgbuild = """
pkgname=__namcap_test_pathdepends
pkgver=1.0
pkgrel=1
pkgdesc="A package"
arch=('i686' 'x86_64')
url="http://www.example.com/"
license=('GPL')
depends=('glibc')
source=()
options=(!purge !zipman)
build() {
  true
}
package() {
  # dconf-needed-for-glib-schemas
  mkdir -p "${pkgdir}/usr/share/glib-2.0/schemas"
  touch "${pkgdir}/usr/share/glib-2.0/schemas/org.test.gschema.xml"

  # glib2-needed-for-gio-modules
  mkdir -p "${pkgdir}/usr/lib/gio/modules"
  touch "${pkgdir}/usr/lib/gio/modules/something.so"

  # hicolor-icon-theme-needed-for-hicolor-dir
  mkdir -p "${pkgdir}/usr/share/icons/hicolor/64x64/apps"
  touch "${pkgdir}/usr/share/icons/hicolor/64x64/apps/example.png"

  # shared-mime-info-needed
  mkdir -p "${pkgdir}/usr/share/mime/text"
  touch "${pkgdir}/usr/share/mime/text/example.xml"
}
"""

	def test_pathdepends_exists(self):
		pkgfile = "__namcap_test_pathdepends-1.0-1-%(arch)s.pkg.tar" % { "arch": self.arch }
		with open(os.path.join(self.tmpdir, "PKGBUILD"), "w") as f:
			f.write(self.pkgbuild)
		self.run_makepkg()
		pkg, r = self.run_rule_on_tarball(
				os.path.join(self.tmpdir, pkgfile),
				Namcap.rules.pathdepends.PathDependsRule
				)
		self.assertEqual(pkg.detected_deps,
				{'dconf': [('dconf-needed-for-glib-schemas', ())],
				 'glib2': [('glib2-needed-for-gio-modules', ())],
				 'hicolor-icon-theme': [('hicolor-icon-theme-needed-for-hicolor-dir', ())],
				 'shared-mime-info': [('shared-mime-info-needed', ())],
				})
		self.assertEqual(r.errors, [])
		self.assertEqual(r.warnings, [])
		self.assertEqual(r.infos, [])


# vim: set ts=4 sw=4 noet:




#
# namcap rules - rpath
# Copyright (C) 2009 Abhishek Dasgupta <abhidg@gmail.com>
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

from Namcap.util import is_elf
from Namcap.ruleclass import *

from elftools.elf.elffile import ELFFile
from elftools.elf.dynamic import DynamicSection

allowed = ['/usr/lib', '/usr/lib32', '/lib', '$ORIGIN', '${ORIGIN}']
allowed_toplevels = [s + '/' for s in allowed]
warn = ['/usr/local/lib']

def get_rpaths(fileobj):
	elffile = ELFFile(fileobj)
	for section in elffile.iter_sections():
		if not isinstance(section, DynamicSection):
			continue
		for tag in section.iter_tags():
			if tag.entry.d_tag != 'DT_RPATH':
				continue
			rpaths = tag.rpath
			rpaths = rpaths.split(':')
			for path in rpaths:
				yield path

class package(TarballRule):
	name = "rpath"
	description = "Verifies correct and secure RPATH for files."
	def analyze(self, pkginfo, tar):
		for entry in tar:
			if not entry.isfile():
				continue

			# is it an ELF file ?
			fileobj = tar.extractfile(entry)
			if not is_elf(fileobj):
				continue

			for path in get_rpaths(fileobj):
				path_ok = path in allowed
				for allowed_toplevel in allowed_toplevels:
					if path.startswith(allowed_toplevel):
						path_ok = True

				if not path_ok:
					self.errors.append(("insecure-rpath %s %s",
						(path, entry.name)))
					break
				if path in warn and entry.name not in insecure_rpaths:
					self.warnings.append(("insecure-rpath %s %s",
						(path, entry.name)))


# vim: set ts=4 sw=4 noet:



# -*- coding: utf-8 -*-
# 
# namcap - tags
# Copyright (C) 2003-2009 Jason Chu <jason@archlinux.org>
# Copyright (c) 2011 Rémy Oudompheng <remy@archlinux.org>
# 
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
# 

import os

tags = {}

DEFAULT_TAGS = "/usr/share/namcap/namcap-tags"

def load_tags(filename = None, machine = False):
	"Loads tags from the given filename"
	global tags
	tags = {}
	if filename is None:
		filename = DEFAULT_TAGS

	f = open(filename)
	for line in f:
		line = line.strip()
		if line.startswith("#") or line == "":
			continue
		machinetag, humantag = line.split("::")
		machinetag = machinetag.strip()
		humantag = humantag.strip()

		if machine:
			tags[machinetag] = machinetag
		else:
			tags[machinetag] = humantag

def format_message(msg):
	"""
	Formats a tuple (tag, data)
	"""
	tag, data = msg
	return (tags[tag] % data)

# Try to load tags by default
if os.path.exists(DEFAULT_TAGS):
	load_tags(DEFAULT_TAGS)
elif os.path.exists("namcap-tags"):
	load_tags("namcap-tags")


# vim: set ts=4 sw=4 noet:



#
# namcap rules - elffiles
# Copyright (C) 2009 Hugo Doria <hugo@archlinux.org>
# Copyright (C) 2011 Rémy Oudompheng <remy@archlinux.org>
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

from elftools.elf.elffile import ELFFile
from elftools.elf.dynamic import DynamicSection
from elftools.elf.sections import SymbolTableSection

from Namcap.util import is_elf
from Namcap.ruleclass import *

# Valid directories for ELF files
valid_dirs = ['bin/', 'sbin/', 'usr/bin/', 'usr/sbin/', 'lib/',
		'usr/lib/', 'usr/lib32/']
# Questionable directories for ELF files
# (Suppresses some output spam.)
questionable_dirs = ['opt/']

class ELFPaths(TarballRule):
	name = "elfpaths"
	description = "Check about ELF files outside some standard paths."
	def analyze(self, pkginfo, tar):
		invalid_elffiles = []
		questionable_elffiles = []

		for entry in tar:
			# is it a regular file ?
			if not entry.isfile():
				continue
			# is it outside standard binary dirs ?
			in_std_dirs = any(entry.name.startswith(d) for d in valid_dirs)
			in_que_dirs = any(entry.name.startswith(d) for d in questionable_dirs)

			if in_std_dirs:
				continue
			# is it an ELF file ?
			f = tar.extractfile(entry)
			if is_elf(f):
				if in_que_dirs:
					questionable_elffiles.append(entry.name)
				else:
					invalid_elffiles.append(entry.name)

		que_elfdirs = [d for d in questionable_dirs if any(f.startswith(d) for f in questionable_elffiles)]
		self.errors = [("elffile-not-in-allowed-dirs %s", i)
				for i in invalid_elffiles]
		self.errors.extend(("elffile-in-questionable-dirs %s", i)
				for i in que_elfdirs)
		self.infos = [("elffile-not-in-allowed-dirs %s", i)
				for i in questionable_elffiles]


class ELFTextRelocationRule(TarballRule):
	"""
	Check for text relocations in ELF files.
	"""

	name = "elftextrel"
	description = "Check for text relocations in ELF files."

	def analyze(self, pkginfo, tar):
		files_with_textrel = []

		for entry in tar:
			if not entry.isfile():
				continue
			fp = tar.extractfile(entry)
			if not is_elf(fp):
				continue
			elffile = ELFFile(fp)
			for section in elffile.iter_sections():
				if not isinstance(section, DynamicSection):
					continue
				for tag in section.iter_tags():
					if tag.entry.d_tag == 'DT_TEXTREL':
						files_with_textrel.append(entry.name)

		if files_with_textrel:
			self.warnings = [("elffile-with-textrel %s", i)
					for i in files_with_textrel]

class ELFExecStackRule(TarballRule):
	"""
	Check for executable stacks in ELF files.

	Introduced by FS#26458. Uses pyelftools to read the GNU_STACK
	program header and ensure it does not have the executable bit
	set.
	"""

	name = "elfexecstack"
	description = "Check for executable stacks in ELF files."

	def analyze(self, pkginfo, tar):
		exec_stacks = []

		for entry in tar:
			if not entry.isfile():
				continue
			fp = tar.extractfile(entry)
			if not is_elf(fp):
				continue
			elffile = ELFFile(fp)
			for segment in elffile.iter_segments():
				if segment['p_type'] != 'PT_GNU_STACK':
					continue

				mode = segment['p_flags']
				if mode & 1:
					exec_stacks.append(entry.name)

		if exec_stacks:
			self.warnings = [("elffile-with-execstack %s", i)
					for i in exec_stacks]

class ELFGnuRelroRule(TarballRule):
	"""
	Check for read-only relocation in ELF files.

	Introduced by FS#26435. Uses pyelftools to check for GNU_RELRO.
	"""

	name = "elfgnurelro"
	description = "Check for FULL RELRO in ELF files."

	def has_bind_now(self, elffile):
		for section in elffile.iter_sections():
			if not isinstance(section, DynamicSection):
				continue
			if any(tag.entry.d_tag == 'DT_BIND_NOW' for tag in section.iter_tags()):
				return True
		return False

	def analyze(self, pkginfo, tar):
		missing_relro = []

		for entry in tar:
			if not entry.isfile():
				continue
			fp = tar.extractfile(entry)
			if not is_elf(fp):
				continue
			elffile = ELFFile(fp)
			if any(seg['p_type'] == 'PT_GNU_RELRO' for seg in elffile.iter_segments()):
				if self.has_bind_now(elffile):
					continue

			missing_relro.append(entry.name)

		if missing_relro:
			self.warnings = [("elffile-without-relro %s", i)
					for i in missing_relro]

class ELFUnstrippedRule(TarballRule):
	"""
	Checks for unstripped ELF files. Uses pyelftools to check if
	.symtab exists.

	Introduced by FS#27485.
	"""

	name = "elfunstripped"
	description = "Check for unstripped ELF files."

	def analyze(self, pkginfo, tar):
		unstripped_binaries = []

		for entry in tar:
			if not entry.isfile():
				continue
			fp = tar.extractfile(entry)
			if not is_elf(fp):
				continue
			elffile = ELFFile(fp)
			for section in elffile.iter_sections():
				if not isinstance(section, SymbolTableSection):
					continue

				if section['sh_entsize'] == 0:
					continue

				if section.name == '.symtab':
					unstripped_binaries.append(entry.name)
		if unstripped_binaries:
			self.warnings = [("elffile-unstripped %s", i)
					for i in unstripped_binaries]

class NoPIERule(TarballRule):
	"""
	Checks for no PIE ELF files.
	"""

	name = "elfnopie"
	description = "Check for no PIE ELF files."

	def has_dt_debug(self, elffile):
		for section in elffile.iter_sections():
			if not isinstance(section, DynamicSection):
				continue
			if any(tag.entry.d_tag == 'DT_DEBUG' for tag in section.iter_tags()):
				return True
		return False

	def analyze(self, pkginfo, tar):
		nopie_binaries = []

		for entry in tar:
			if not entry.isfile():
				continue
			fp = tar.extractfile(entry)
			if not is_elf(fp):
				continue
			elffile = ELFFile(fp)
			if elffile.header['e_type'] != 'ET_DYN' or not self.has_dt_debug(elffile):
				nopie_binaries.append(entry.name)

		if nopie_binaries:
			self.warnings = [("elffile-nopie %s", i) for i in nopie_binaries]


# vim: set ts=4 sw=4 noet:



#
# namcap rules - unusedsodepends
# Copyright (C) 2009 Abhishek Dasgupta <abhidg@gmail.com>
# Copyright (C) 2013 Jan Alexander Steffens <jan.steffens@gmail.com>
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

import os, subprocess, re
import tempfile
from Namcap.util import is_elf
from Namcap.ruleclass import *

libre = re.compile('^\t(/.*)')
lddfail = re.compile('^\tnot a dynamic executable')

def get_unused_sodepends(filename):
	p = subprocess.Popen(["ldd", "-r", "-u", filename],
		env={'LANG': 'C'},
		stdout=subprocess.PIPE,
		stderr=subprocess.PIPE)
	var = p.communicate()
	if p.returncode == 0:
		return
	for j in var[0].decode('ascii').splitlines():
		# Don't raise an error, as the executable might be a valid ELF file,
		# just not a dynamically linked one
		n = lddfail.search(j)
		if n is not None:
			return

		n = libre.search(j)
		if n is not None:
			yield n.group(1)

class package(TarballRule):
	name = "unusedsodepends"
	description = "Checks for unused dependencies caused by linked shared libraries"
	def analyze(self, pkginfo, tar):
		for entry in tar:
			if not entry.isfile():
				continue

			# is it an ELF file ?
			f = tar.extractfile(entry)
			if not is_elf(f):
				f.close()
				continue
			elf = f.read()
			f.close()

			# write it to a temporary file
			f = tempfile.NamedTemporaryFile(delete = False)
			f.write(elf)
			f.close()

			os.chmod(f.name, 0o755)

			for lib in get_unused_sodepends(f.name):
				self.warnings.append(("unused-sodepend %s %s", (lib, entry.name)))

			os.unlink(f.name)

# vim: set ts=4 sw=4 noet:



# -*- coding: utf-8 -*-
#
# namcap tests - extravars
# Copyright (C) 2011 Rémy Oudompheng <remy@archlinux.org>
# 
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
#   USA
# 

import os
from Namcap.tests.pkgbuild_test import PkgbuildTest
import Namcap.rules.extravars

class ExtravarsTest(PkgbuildTest):
	pkgbuild1 = """
# Maintainer: Arch Linux <archlinux@example.com>
# Contributor: Arch Linux <archlinux@example.com>

pkgname=mypackage
pkgver=1.0
pkgrel=1
pkgdesc="A package"
arch=('i686' 'x86_64')
url="http://www.example.com/"
license='GPL'
depends=('glibc')
depends+=('glib2')
optdepends=('gtk2')
options=('!libtool')
source=(ftp://ftp.example.com/pub/mypackage-0.1.tar.gz)
md5sums=('abcdefabcdef12345678901234567890')

mycustomvar="something"
_legalvar="something else"

build() {
  cd "${srcdir}"/${pkgname}-${pkgver}
  ./configure --prefix=/usr \\
	--libdir=something
  make
  hello=world make something
}

package() {
  cd "${srcdir}"/${pkgname}-${pkgver}
  make DESTDIR="${pkgdir}" install
}
"""
	# now a more tricky example
	pkgbuild2 = """
# Maintainer: Arch Linux <archlinux@example.com>
# Contributor: Arch Linux <archlinux@example.com>

pkgname=mypackage
pkgver=1.0
pkgrel=1
pkgdesc="A pac
kage=1 package"
arch=('i686' 'x86_64')
url="http://www.example.com/"
license=('GPL')
depends=('glibc')
depends+=('glib2'
glibc=2.12
)
depends_i686=('zlib')
depends_x86_64=('lib32-zlib')
optdepends=('gtk2')
options=('!libtool')
source=(ftp://ftp.example.com/pub/mypackage-0.1.tar.gz)
validpgpkeys=('D7EDD665AB6F255E2EB9BA2702FA998FBAD6674A')
md5sums=('abcdefabcdef12345678901234567890')

build() {
  cd "${srcdir}"/${pkgname}-${pkgver}
  ./configure --prefix=/usr \\
	--libdir=something
  _variable=value make; mycustomvar="something"
  hello=world make something
  echo 's/a/b/
        b=c
        s/c=d/'
}

package() {
  cd "${srcdir}"/${pkgname}-${pkgver}
  make DESTDIR="${pkgdir}" install
}
"""

	test_valid = PkgbuildTest.valid_tests

	def preSetUp(self):
		self.rule = Namcap.rules.extravars.package

	def test_example1(self):
		"PKGBUILD with custom variables without underscore"
		os.putenv("environment_pollution", "yes")
		r = self.run_on_pkg(self.pkgbuild1)
		self.assertEqual(r.errors, [])
		self.assertEqual(set(r.warnings), set([
			("extra-var-begins-without-underscore %s", "mycustomvar"),
		]))
		self.assertEqual(r.infos, [])

	def test_example2(self):
		"a tricky PKGBUILD with custom variables without underscore"
		r = self.run_on_pkg(self.pkgbuild2)
		self.assertEqual(r.errors, [])
		self.assertEqual(r.warnings, [])
		self.assertEqual(r.infos, [])


# vim: set ts=4 sw=4 noet:



#
# namcap rules - py_mtime
# Copyright (C) 2013 Kyle Keen <keener@gmail.com>
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

"""
Check for py timestamps that are ahead of pyc/pyo timestamps
"""

import os
from Namcap.package import load_mtree
from Namcap.ruleclass import *

def _quick_filter(names):
	"can this package be skipped outright"
	if not names:
		return True
	found_py  = any(n.endswith('.py')  for n in names)
	found_pyc = any(n.endswith('.pyc') for n in names)
	found_pyo = any(n.endswith('.pyo') for n in names)
	if found_py and found_pyc:
		return False
	if found_py and found_pyo:
		return False
	return True

def _tar_timestamps(tar):
	"takes a tar object"
	return dict((m.name, m.mtime) for m in tar.getmembers())

def _mtree_timestamps(tar):
	"takes a tar object"
	return dict((h, a['time']) for h,a in load_mtree(tar) if 'time' in a)

def _generic_timestamps(tar):
	"works for mtree and tar"
	if '.MTREE' in tar.getnames():
		return _mtree_timestamps(tar)
	return _tar_timestamps(tar)

def _try_mtree(tar):
	"returns True if good, False if bad, None if N/A"
	if '.MTREE' not in tar.getnames():
		return None
	stamps = _mtree_timestamps(tar)
	if _quick_filter(stamps.keys()):
		return True
	return not _mtime_filter(stamps)

def _try_tar(tar):
	"returns True if good, False if bad"
	names = tar.getnames()
	if _quick_filter(names):
		return True
	mtimes = _tar_timestamps(tar)
	return not _mtime_filter(mtimes)

def _split_all(path):
	"like os.path.split but splits every directory"
	p2 = path
	dirs = []
	while p2 and p2 != '/':
		p2,p3 = os.path.split(p2)
		dirs.insert(0, p3)
	#dirs.insert(0, '/')
	return dirs

def _source_py(path):
	"given a pyc/pyo, return the source path"
	if not path.endswith('.pyc') and not path.endswith('.pyo'):
		return None
	path = path[:-1]
	# handle py2
	if '__pycache__' not in path:
		return path
	# handle py3
	splitup = _split_all(path)
	if splitup[-2] != '__pycache__':
		return None
	splitup.pop(-2)
	f = splitup[-1]
	f = f.split('.')
	f.pop(-2)
	splitup[-1] = '.'.join(f)
	return os.path.join(*splitup)

def _mtime_filter(mtimes):
	"return list of bad py file names"
	bad = []
	for name, mt2 in mtimes.items():
		if not name.endswith('.pyc') and not name.endswith('.pyo'):
			continue
		source_name = _source_py(name)
		if source_name not in mtimes:
			continue
		mt1 = mtimes[source_name]
		if mt1 > mt2:
			bad.append(source_name)
	return bad

class package(TarballRule):
	name = "py_mtime"
	description = "Check for py timestamps that are ahead of pyc/pyo timestamps"
	def analyze(self, pkginfo, tar):
		mtree_status = _try_mtree(tar)
		tar_status = _try_tar(tar)
		if mtree_status == False and tar_status:
			# mtree only
			self.warnings = [('py-mtime-mtree-warning', ())]
		elif not tar_status:
			# tar or both
			self.errors = [('py-mtime-tar-error', ())]
		self.infos = [('py-mtime-file-name %s', f[1:]) for f in _mtime_filter(_generic_timestamps(tar))]

# vim: set ts=4 sw=4 noet:



#
# namcap rules - utility functions
# Copyright (C) 2009 Dan McGee <dan@archlinux.org>
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

import re

def _file_has_magic(fileobj, magic_bytes):
	length = len(magic_bytes)
	magic = fileobj.read(length)
	fileobj.seek(0)
	return magic == magic_bytes

def is_elf(fileobj):
	"Take file object, peek at the magic bytes to check if ELF file."
	return _file_has_magic(fileobj, b"\x7fELF")

def is_static(fileobj):
	"Take file object, peek at the magic bytes to check if static lib."
	return _file_has_magic(fileobj, b"!<arch>\n")

def is_script(fileobj):
	"Take file object, peek at the magic bytes to check if script."
	return _file_has_magic(fileobj, b"#!")

def is_java(fileobj):
	"Take file object, peek at the magic bytes to check if class file."
	return _file_has_magic(fileobj, b"\xCA\xFE\xBA\xBE")

def script_type(fileobj):
	firstline = fileobj.readline()
	fileobj.seek(0)
	try:
		firstline = firstline.decode('utf-8', 'strict')
	except UnicodeDecodeError:
		return None
	if not firstline:
		return None
	script = re.compile('#!.*/(.*)')
	m = script.match(firstline)
	if m is None:
		return None
	cmd = m.group(1).split()
	name = cmd[0]
	if name == 'env':
		name = cmd[1]
	return name

# vim: set ts=4 sw=4 noet:



# 
# namcap rules - makepkgfunctions
# Copyright (C) 2017 Kyle Keen <keenerd@gmail.com>
# 
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
# 

import re
from Namcap.ruleclass import *

class package(PkgbuildRule):
	name = "makepkgfunctions"
	description = "Looks for calls to makepkg functionality"
	def analyze(self, pkginfo, tar):
		bad_calls = ['msg', 'msg2', 'warning', 'error', 'plain']
		regex = re.compile('^\s+(%s) ' % '|'.join(bad_calls))
		hits = set()
		for i in pkginfo.pkgbuild:
			if regex.match(i):
				call = regex.match(i).group(1)
				hits.add(call)
		for i in hits:
			self.warnings.append(("makepkg-function-used %s", i))

# vim: set ts=4 sw=4 noet:



# -*- coding: utf-8 -*-
#
# namcap rules - sodepends
# Copyright (C) 2003-2009 Jason Chu <jason@archlinux.org>
# Copyright (C) 2011 Rémy Oudompheng <remy@archlinux.org>
# 
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
# 

"""Checks dependencies resulting from linking of shared libraries."""

from collections import defaultdict
import re
import os
import subprocess
import Namcap.package
from Namcap.ruleclass import *
from Namcap.util import is_elf
from Namcap.rules.rpath import get_rpaths

from elftools.elf.elffile import ELFFile
from elftools.elf.dynamic import DynamicSection

libcache = {'i686': {}, 'x86-64': {}}

def scanlibs(fileobj, filename, custom_libs):
	"""
	Find shared libraries in a file-like binary object

	If it depends on a library, store that library's path.

	returns: a dictionary { library => set(ELF files using that library) }
	"""

	if not is_elf(fileobj):
		return {}

	elffile = ELFFile(fileobj)
	sharedlibs = defaultdict(set)
	for section in elffile.iter_sections():
		if not isinstance(section, DynamicSection):
			continue
		for tag in section.iter_tags():
			# DT_NEEDED means shared library
			if tag.entry.d_tag != 'DT_NEEDED':
				continue
			bitsize = elffile.elfclass
			architecture = {32:'i686', 64:'x86-64'}[bitsize]
			libname = tag.needed
			if libname in custom_libs:
				sharedlibs[custom_libs[libname][1:]].add(filename)
				continue
			try:
				libpath = os.path.abspath(
						libcache[architecture][libname])[1:]
				sharedlibs[libpath].add(filename)
			except KeyError:
				# We didn't know about the library, so add it for fail later
				sharedlibs[libname].add(filename)
	return sharedlibs

def finddepends(liblist):
	"""
	Find packages owning a list of libraries

	Returns:
	  dependlist -- a dictionary { package => set(libraries) }
	  orphans -- the list of libraries without owners
	"""
	dependlist = defaultdict(set)

	somatches = {}
	actualpath = {}

	knownlibs = set(liblist)
	foundlibs = set()

	actualpath = dict((j, os.path.realpath('/' + j)[1:]) for j in knownlibs)

	# Sometimes packages don't include all so .so, .so.1, .so.1.13, .so.1.13.19 files
	# They rely on ldconfig to create all the symlinks
	# So we will strip off the matching part of the files and use this regexp to match the rest
	so_end = re.compile('(\.\d+)*')
	# Whether we should even look at a particular file
	is_so = re.compile('\.so')

	for pkg in Namcap.package.get_installed_packages():
		for j, fsize, fmode in pkg.files:
			if not is_so.search(j):
				continue

			for k in knownlibs:
				# File must be an exact match or have the right .so ending numbers
				# i.e. gpm includes libgpm.so and libgpm.so.1.19.0, but everything links to libgpm.so.1
				# We compare find libgpm.so.1.19.0 startswith libgpm.so.1 and .19.0 matches the regexp
				if j == actualpath[k] or (j.startswith(actualpath[k]) and so_end.match(j[len(actualpath[k]):])):
					dependlist[pkg.name].add(k)
					foundlibs.add(k)

	orphans = list(knownlibs - foundlibs)
	return dependlist, orphans

def filllibcache():
	var = subprocess.Popen('ldconfig -p', 
			env = {"LANG": "C"},
			shell=True,
			stdout=subprocess.PIPE,
			stderr=subprocess.PIPE).communicate()
	libline = re.compile('\s*(.*) \((.*)\) => (.*)')
	for j in var[0].decode('ascii').splitlines():
		g = libline.match(j)
		if g != None:
			if g.group(2).startswith('libc6,x86-64'):
				libcache['x86-64'][g.group(1)] = g.group(3)
			else:
				# TODO: This is bogus; what do non x86-architectures print?
				libcache['i686'][g.group(1)] = g.group(3)


class SharedLibsRule(TarballRule):
	name = "sodepends"
	description = "Checks dependencies caused by linked shared libraries"
	def analyze(self, pkginfo, tar):
		liblist = {}
		dependlist = {}
		filllibcache()
		os.environ['LC_ALL'] = 'C'
		pkg_so_files = ['/' + n for n in tar.getnames() if '.so' in n]

		for entry in tar:
			if not entry.isfile():
				continue
			f = tar.extractfile(entry)
			# find anything that could be rpath related
			rpath_files = {}
			if is_elf(f):
				rpaths = list(get_rpaths(f))
				f.seek(0)
				for n in pkg_so_files:
					if any(n.startswith(rp) for rp in rpaths):
						rpath_files[os.path.basename(n)] = n
			liblist.update(scanlibs(f, entry.name, rpath_files))
			f.close()

		# Ldd all the files and find all the link and script dependencies
		dependlist, orphans = finddepends(liblist)

		# Handle "no package associated" errors
		self.warnings.extend([("library-no-package-associated %s", i)
			for i in orphans])

		# Print link-level deps
		for pkg, libraries in dependlist.items():
			if isinstance(libraries, set):
				files = list(libraries)
				needing = set().union(*[liblist[lib] for lib in libraries])
				reasons = pkginfo.detected_deps.setdefault(pkg, [])
				reasons.append((
					"libraries-needed %s %s",
					(str(files), str(list(needing)))
					))
				self.infos.append(("link-level-dependence %s in %s", (pkg, str(files))))

		# Check for packages in testing
		for i in dependlist.keys():
			p = Namcap.package.load_testing_package(i)
			q = Namcap.package.load_from_db(i)
			if p is not None and q is not None and p["version"] == q["version"] :
				self.warnings.append(("dependency-is-testing-release %s", i))

# vim: set ts=4 sw=4 noet:



# 
# namcap rules - fhs
# Copyright (C) 2003-2009 Jason Chu <jason@archlinux.org>
# 
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
# 

import os, re
from Namcap.ruleclass import *

class FHSRule(TarballRule):
	name = "directoryname"
	description = "Checks for standard directories."
	def analyze(self, pkginfo, tar):
		valid_paths = [
				'etc/', 'opt/',
				'lib/modules',
				'usr/bin/', 'usr/include/', 'usr/lib/', 'usr/lib32/',
				'usr/sbin/', 'usr/share/', 'usr/src/',
				'var/cache/', 'var/lib/', 'var/log/', 'var/opt/',
				'var/spool/', 'var/state/',
				'.PKGINFO', '.INSTALL', '.CHANGELOG', '.MTREE', '.BUILDINFO',
		]
		forbidden_paths = [
				'tmp/', 'var/tmp/',
				'run/', 'var/run/',
				'var/lock/'
		]
		custom_valid = {
			'^mingw-': ['usr/x86_64-w64-mingw32/lib/',
			            'usr/x86_64-w64-mingw32/bin/',
						'usr/x86_64-w64-mingw32/include/',
						'usr/i686-w64-mingw32/lib/',
						'usr/i686-w64-mingw32/bin/',
						'usr/i686-w64-mingw32/include/'],
			}
		for pattern in custom_valid:
			if re.search(pattern, pkginfo['name']):
				valid_paths.extend(custom_valid[pattern])
		for entry in tar.getmembers():
			name = os.path.normpath(entry.name)
			if entry.isdir():
				name += '/'

			# check for files in wrong dirs, directory itself will be
			# catched by emptydirs rule
			if name in forbidden_paths:
				continue
			bad_dirs = (name.startswith(dirname) for dirname in forbidden_paths)
			if any(bad_dirs):
				self.errors.append(('file-in-temporary-dir %s',	name))
				continue

			# matches directory names or parent directories
			good_dirs = (name.startswith(dirname) or dirname.startswith(name)
				for dirname in valid_paths)
			if not any(good_dirs):
				self.warnings.append(("file-in-non-standard-dir %s", name))

class FHSManpagesRule(TarballRule):
	name = "fhs-manpages"
	description = "Verifies correct installation of man pages"
	def analyze(self, pkginfo, tar):
		gooddir = 'usr/share/man'
		bad_dir = 'usr/man'
		for i in tar.getmembers():
			if not i.isfile():
				continue
			if i.name.startswith(gooddir):
				continue
			if i.name.startswith(bad_dir):
				self.errors.append(("non-fhs-man-page %s", i.name))
				continue
			#Check everything else to see if it has a 'man' path component
			if "man" in i.name.split(os.sep):
				self.warnings.append(("potential-non-fhs-man-page %s", i.name))

class FHSInfoPagesRule(TarballRule):
	name = "fhs-infopages"
	description = "Verifies correct installation of info pages"
	def analyze(self, pkginfo, tar):
		for i in tar.getmembers():
			if not i.isfile():
				continue
			if i.name.startswith('usr/share/info'):
				continue
			if i.name.startswith('usr/info'):
				self.errors.append(("non-fhs-info-page %s", i.name))
				continue
			if "info" in i.name.split(os.sep):
				self.warnings.append(("potential-non-fhs-info-page %s", i.name))

class RubyPathsRule(TarballRule):
	name = "rubypaths"
	description = "Verifies correct usage of folders by ruby packages"
	def analyze(self, pkginfo, tar):
		for i in tar.getmembers():
			if i.name.startswith('usr/lib/ruby/site_ruby'):
				self.warnings.append(("site-ruby", ()))
				return

# vim: set ts=4 sw=4 noet:



# -*- coding: utf-8 -*-
#
# namcap tests - carch
# Copyright (C) 2011 Rémy Oudompheng <remy@archlinux.org>
# 
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
#   USA
# 

from Namcap.tests.pkgbuild_test import PkgbuildTest
import Namcap.rules.carch as module

class NamcapSpecialArchTest(PkgbuildTest):
	pkgbuild1 = """
pkgname=mypackage
pkgver=1.0
pkgrel=1
pkgdesc="A package"
arch=('i686')

build() {
  cd "${srcdir}"/${pkgname}-${pkgver}
  [[ $CARCH == x86_64 ]] && CFLAGS+="-m64"
  [ '$CARCH' = 'i686' ] && CFLAGS+="-m32"
  ./configure --prefix=/usr
  make
}
"""

	pkgbuild2 = """
pkgname=mypackage
pkgver=1.0
pkgrel=1
pkgdesc="A package"
arch=('i686')

build() {
  cd "${srcdir}"/${pkgname}-${pkgver}
  [[ $CARCH == x86_64 ]] && CFLAGS+="-m64"
  [ '$CARCH' = 'i686' ] && CFLAGS+="-m32"
  ./configure --prefix=/usr
  make
}

package() {
  cd "${srcdir}"/${pkgname}-${pkgver}
  ./configure --prefix=/usr
  make DESTDIR="${pkgdir}" install
  cp foobar /usr/lib/i686/pkg/
}
"""

	test_valid = PkgbuildTest.valid_tests

	def preSetUp(self):
		self.rule = module.package

	def test_example1(self):
		r = self.run_on_pkg(self.pkgbuild1)
		self.assertEqual(r.errors, [])
		self.assertEqual(r.warnings, [])
		self.assertEqual(r.infos, [])

	def test_example2(self):
		r = self.run_on_pkg(self.pkgbuild2)
		self.assertEqual(r.errors, [])
		self.assertEqual(r.warnings, [("specific-host-type-used %s", "i686")])
		self.assertEqual(r.infos, [])


# vim: set ts=4 sw=4 noet:
